%%%
%%% goes back to  parser.prec.tucr19.clingo
%%%
% from lexicon:
%
% word_entry/2
%   links words in lexicon to lexicon entries (words are strings)
% entry_category/2
%   gives one (possibly complex) category for each lexicon entry
%   categories are given in functional style:
%   rfunc(C1,C2) denotes a CCG functor "C1/C2" which
%     has type C1 and takes an argument C2 to the right
%   lfunc(C1,C2) denotes a CCG functor "C1\C2" which
%     has type C1 and takes an argument C2 to the left
%   (basic categories are strings)

% from input:
%
% word_at/2
%   gives word at position in sentence
%   (word is given as string)
%   (position is given as integer)
% adjacent/2
%   defines adjacency between sentence positions (ordered)

%
% the parser
%
%
% parser configuration
%


% conjunctions:
% conjunction_category/1
conjunction_category("conj").
conjunction_category("Conj").



% backwards cross substitution licensing
% licensed for Y = "VP-like"
% (see www.ling.umd.edu/~alxndrw/CGReadings/steedman-dowty-1.pdf pg.5)
license_bwd_x_subst(lfunc("S","NP")).
license_bwd_x_subst(rfunc(lfunc("S","NP"),"NP")).
license_bwd_x_subst(rfunc(rfunc(lfunc("S","NP"),"NP"),"NP")).

% punctuation categories
%
% punctuation categories
%
% define "S\S" for "." (full stop)
word_entry(".",fullstop).
entry_category(fullstop,lfunc("S","S")).
% do not allow any other category than "S\S" for "." (full stop)
:- word_at(".",P), not position_category(P,lfunc("S","S"),0).
% forbid that "the" is "NP" initally
%:- word_at("the",P), position_category(P,"NP",0).

% see Julia Hockenmaier's thesis pg 34 and 70
% define "(S/S)\NP" for "," (NP extraction left)
word_entry(",",npCommaExtractionL).
entry_category(npCommaExtractionL,lfunc(rfunc("S","S"),"NP")).
% define "(S\S)/NP" for "," (NP extraction right)
word_entry(",",npCommaExtractionR).
entry_category(npCommaExtractionR,rfunc(lfunc("S","S"),"NP")).
% define "((S\NP)\(S\NP))/NP" for "," (NP extraction right)
word_entry(",",npCommaExtractionR).
entry_category(npCommaExtractionR,rfunc(lfunc("S","S"),"NP")).

%% define "(NP\NP)/NP" for ","
%% (NP adjunct to NP, separated by "," as in "Bruce, the brute")
%word_entry(",",commaNPAdjunct).
%entry_category(commaNPAdjunct,rfunc(lfunc("NP","NP"),"NP")).

% define "S[dcl]\S[dcl]" for ","
word_entry(",",emptyComma).
entry_category(emptyComma,lfunc("S[dcl]","S[dcl]")).

% define conjunction for ","
word_entry(",",commaConj).
entry_category(commaConj,"conj").

%
% goal
%

% one position has category S
% only one position is open
finished(T) :- time(T),
	position(P),
	position_category(P,"S",T),
	1 { position_status_open(P2,T) } 1.
finishedSomewhere :- finished(T).
% must be finished
:- not finishedSomewhere, strict_parse.
%strict_parse.

% alternative: try as hard as possible to parse
minimizationCriterion(P,dummy1) :- position_status_open(P,maxsteps).
%#minimize { position_status_open(P,maxsteps) }.
%#minimize { minimizationCriterion(X,Y) }.

%
% auxiliaries
%

% timepoints
% use "--const maxsteps=N" to override this default via the commandline
#const maxsteps = 6.
time(0..maxsteps).
actiontime(0..maxsteps-1).
t_tminus1(T,T-1):-time(T), T>0.

% define positions
position(P) :- word_at(_,P).



% timepoints are used as follows:
% * 0 is the initial timepoint before any rules have been applied
% * actions at T are (guessed and) applied depending on fluents at T
% * fluents at T are calculated from actions applied at T-1 and fluents at T-1
% rationale: this encoding is suitable for iclingo, and an encoding where action
% at T determines fluents at T+1 we cannot use iclingo

% fluent: position_category/3 (Position, Category, Time)
%    the category of a position (can be changed by rules)
% fluent: position_status_open/2 (Position, Time)
%    whether a position can still be used ("open") or has been
%    dealt with by a prior rule application ("closed")
%    if position_status_open holds then that position is open,
%    i.e., rules still can apply to that position
% fluent: position_affected/2(Position,Time)
%    whether position was affected by action at time Time
%    (not relevant for closed positions, true for initial conditions)
%    (this is used for normalizing the parse/reducing possibilities,
%     by only allowing actions to apply to positions where one
%     position has recently been changed (this enforces that the parser
%     never "waits for applying a rule")
%    (as actions might not change the category, we cannot implictly
%     calculate this fluent)
% implicit fluent: position_adjacent/3 (PositionLeft, PositionRight, Time)
%    whether two positions are adjacent
%    (closing positions implicitly rearranges this fluent)
% implicit fluent: position_last_affected/3 (Position, LastTime, Time)
%    at time Time, when has position been affected last?
%    (initially 0, implicitly updated by position_affected/2)


%%
%% Parser Generic Part
%%

type(unary;binary;ternary).


place_edges(P,T):-place_edges_tag(P,T,TAG).

% if a position is open then we can place at lest one edge from this node to its "parent" 
% in other words at least unary i.e. type raising combinators are applicable
% place_edges  is associated with exectly one TAG which corresponds 
% to the fact that place_edges will either place one, two, or three edges
0{place_edges_tag(P,T,TAG):type(TAG)}1:-position_status_open(P,T), actiontime(T).



%impossible to place a binary or ternary edge if there is some edge placed immediately to the right

%:-place_edges_tag(L,T,binary), position_adjacent(L,R,T), place_edges(R,T).



%impossible to place a  ternary edge if there is some edge placed in one position to the right
%:-place_edges_tag(L,T,ternary), position_conj_adjacent(L,R,T),  place_edges(R,T).

:- -position_status_open(P,T), position_affected(P,T).

% TO ensure preference for "parallel" plans
%
% impossible to perform unary combinator unless it is performed on previously affected position
:-place_edges_tag(P,T,unary), not position_affected(P,T).

% impossible to perform binary/ternary combinator unless it is performed on positions one of which was previously affected 
:-place_edges_tag(P,T,binary), {position_affected(P,T), position_affected(P1,T)}0, position_adjacent(P,P1,T).
:-place_edges_tag(P,T,ternary), {position_affected(P,T), position_affected(P2,T)}0, 
  position_conj_adjacent(P,P2,T).





%
%if there is a unary edge placed then one of the unary action occurs 
1{occ(A,P,T):unary(A)}1:-place_edges_tag(P,T,unary).

%it is impossible for raise to occur unless it happens on the appropriate position
%:- occ(A,P,T), raisingCombinator(A), position_category(P,RaiseFrom,T), not limitRaiseFrom(RaiseFrom).





1{occurs(rule_fwd_raise,P,RaiseRes,T):possibleRaiseRes(rule_fwd_raise,RaiseFrom,RaiseRes)}1 :-occ(rule_fwd_raise,P,T),position_category(P,RaiseFrom,T).
1{occurs(rule_bwd_raise,P,RaiseRes,T):possibleRaiseRes(rule_bwd_raise,RaiseFrom,RaiseRes)}1 :-occ(rule_bwd_raise,P,T),position_category(P,RaiseFrom,T).

possibleRaiseRes(A,RaiseFrom,RaiseRes):-raisingCombinator(A), limitRaise(RaiseFrom,RaiseTo),raiseHelp(A,Raisefrom,RaiseTo,RaiseRes).

%:-occurs(A,P,RaiseRes,T), position_category(P,RaiseFrom,T), not possibleRaiseRes(A,RaiseFrom,RaiseRes).

1{occurs(rule_unary_type_change,L,Y,T):unary_type_change(X,Y)}1:- occ(rule_unary_type_change,L,T),
	position_category(L,X,T).
%possible_unary_type_change(X):-unary_type_change(X,Y).
%:-occ(rule_unary_type_change,L,T),position_category(L,X,T), not possible_unary_type_change(X). 

position_category(P,RaiseRes,T) :- t_tminus1(T,T1), time(T),
	occurs(A,P,RaiseRes,T1),
    unary(A).





%
%if there is a binary edge placed then one of the binary action occurs 
1{occurs(A,P,T):binary(A)}1:-place_edges_tag(P,T,binary).

%
%if there is a ternary edge placed then one of the ternary actions occurs 
occurs(rule_coordination,P,T):-place_edges_tag(P,T,ternary).


%binary/ternary action cannot occur if its preconditions are unsat
:-occurs(A,L,T), not prec(A,L,T).


%
% inertia
%

% we explicitly modify categories
%   by setting -position_category(P,C,T) and position_category(P,C,T)
position_category(P,C,T) :- t_tminus1(T,T1),
	time(T),  position_category(P,C,T1),
	not position_affected(P,T).

% we explicitly close positions
%   using -position_status_open(P,T)
position_status_open(P,T) :- t_tminus1(T,T1),
	time(T),  position_status_open(P,T1),
	not -position_status_open(P,T).

% we implicitly modify adjacency by closing positions
% so this is not really inertia, but adjacency is derived from position status
% STATIC causal laws: caused F if G ?
position_adjacent(L,R,T) :- t_tminus1(T,T1),
	time(T),  position_adjacent(L,R,T1),
	position_status_open(L,T),
	position_status_open(R,T).

% p1      p2 p3 p4
% -----   --------
% unary   ternary     (and alike situations)
% o       x  o  x     (o open; x closed) open are adjacent now

position_adjacent(L,R,T) :- t_tminus1(T,T1),
	time(T), 
	position_adjacent(L,M,T1), position_adjacent(M,R,T1),
	position_status_open(L,T),
	not position_status_open(M,T),
	position_status_open(R,T).


%
% p1 p2   p3 p4 p5
% -----   --------
% binary  ternary     (and alike situations)
% o  x    x  o  x     (o open; x closed) open are adjacent now

position_adjacent(L,R,T) :- t_tminus1(T,T1),
	time(T), 
	position_adjacent(L,M1,T1), position_adjacent(M1,M2,T1), position_adjacent(M2,R,T1),
	position_status_open(L,T),
	not position_status_open(M1,T),
	not position_status_open(M2,T),
	position_status_open(R,T),
	position_category(M1,CC,0),
    	conjunction_category(CC).



% we implicitly modify position_last_affected by explicitly affecting positions
position_last_affected(P,T,T) :-
	position_affected(P,T).
position_last_affected(P,TLast,T) :- t_tminus1(T,T1),
	position_last_affected(P,TLast,T1),
	position(P), time(T1), time(T),
	not position_affected(P,T).


%
% initial state:
%

% positions have initial categories
%
% each word might have multiple lexicon entries
%
% additionally each position may have multiple tagged
% category entries
%
% one category per position is considered at a time
possible_category_for_position(C,P):-position(P),
									word_at(W,P),
									word_entry(W,E),
									entry_category(E,C).
possible_category_for_position(C,P):-position(P),
									category_tag(P,C,Class),
									use_features,
                  use_category_class(Class).
possible_category_for_position(C,P):-position(P),
									category_tag_nofeatures(P,C,Class),
									not use_features,
                  use_category_class(Class).

%%
%%comma and semicolon may be conjunction categor (generic treatment)
possible_category_for_position("conj",P):-possible_category_for_position(",",P).
possible_category_for_position("conj",P):-possible_category_for_position(";",P).


1{position_category(P,C,0): possible_category_for_position(C,P)}1:-
	position(P).


% positions are open
position_status_open(P,0) :-
	position(P).

% positions have initial adjacency
position_adjacent(PL,PR,0) :-
	adjacent(PL,PR).

% we can apply an action to all initial categories
position_affected(P,0) :-
	position(P).

position_last_affected(P,0,0) :-
	position(P).


%
% REIFIED EFFECTS common to ALL ACTIONS: 
% 1. closes right position (if binary action. i.e., not type raising)
% 
% binary or ternary action occurs
-position_status_open(R,T+1) :- place_edges(L,T), not place_edges_tag(L,T,unary), position_adjacent(L,R,T).
%  ternary action occurs
-position_status_open(R,T+1) :- place_edges_tag(L,T,ternary), position_adjacent(L,M,T), position_adjacent(M,R,T).


% 2. each action that occurs marks its left position as affected
%
position_affected(P,T+1) :- place_edges(P,T).






%
% type raising configuration
%
%
% limitRaise/2 (FromCategory,ToCategory)
%   limits applicability of type raising:
%   to which category raising may apply,
%   and to which category may be raised.
%   if raising shall be possible to any category, use
%   "limitRaise(C1,C2) :- category(C1), category(C2)."

% extra indirection
raiseHelp(rule_fwd_raise,From,To,rfunc(To,lfunc(To,From))) :- limitRaise(From,To).
raiseHelp(rule_bwd_raise,From,To,lfunc(To,rfunc(To,From))) :- limitRaise(From,To).



%%
%% Binary & Ternary prec and effects
%%

prec(A,L,T):-precCat(A,L,T,C).
position_category(L,C,T) :- t_tminus1(T,T1), time(T),
	precCat(A,L,T1,C), occurs(A,L,T1).	







%%%
%%% Specific combinators related
%%%



%%
%% combinator ontology
%%

unary(rule_fwd_raise; rule_bwd_raise; rule_unary_type_change).

binary(rule_fwd_appl;rule_bwd_appl;
                       rule_fwd_comp;rule_bwd_comp;
                       rule_bwd_x_comp;
					   rule_bwd_x_subst;
rule_left_comma;
rule_left_colon;
rule_left_semicolon;
rule_right_comma;
rule_right_colon;
rule_right_semicolon
).

ternary(rule_coordination).

%%
%% Standard binary combinators precCat predicate instantiaton
%%



%
% forward application rule: X/Y Y => X
%
% action: rule_fwd_appl/3 (PositionLeft, PositionRight, Time)

precCat(rule_fwd_appl,L,T,X):-actiontime(T),
						position_adjacent(L,R,T),
    position_category(L,rfunc(X,Y),T),
	position_category(R,Y,T).



%
% backward application rule: Y X\Y => X
%
% action: rule_bwd_appl/3 (PositionLeft, PositionRight, Time)

precCat(rule_bwd_appl,L,T,X):-actiontime(T),
	position_adjacent(L,R,T),		
	position_category(L,Y,T),
	position_category(R,lfunc(X,Y),T).




%
% forward composition rule: X/Y Y/Z  => X/Z
%
% action: rule_fwd_comp/3 (PositionLeft, PositionRight, Time)

precCat(rule_fwd_comp, L,T,rfunc(X,Z)) :- actiontime(T),  
	position_adjacent(L,R,T),				 
	position_category(L,rfunc(X,Y),T),
	position_category(R,rfunc(Y,Z),T).



%
% backward composition rule: Y\Z X\Y => X\Z
%
% action: rule_bwd_comp/3 (PositionLeft, PositionRight, Time)

precCat(rule_bwd_comp,L,T,lfunc(X,Z))  :- actiontime(T),
	position_adjacent(L,R,T),				
	position_category(L,lfunc(Y,Z),T),
	position_category(R,lfunc(X,Y),T).




%
% backward cross composition rule: Y/Z X\Y => X/Z
%
% action: rule_bwd_x_comp/3 (PositionLeft, PositionRight, Time)

precCat(rule_bwd_x_comp,L,T,rfunc(X,Z)) :- actiontime(T),
    position_adjacent(L,R,T),
	position_category(L,rfunc(Y,Z),T),
	position_category(R,lfunc(X,Y),T).



%
% backward cross substitution rule: Y/Z (X\Y)/Z => X/Z
%
% action: rule_bwd_x_subst/3 (PositionLeft, PositionRight, Time)
%
precCat(rule_bwd_x_subst,L,T,rfunc(X,Z)) :- actiontime(T),	
    position_adjacent(L,R,T),
	position_category(L,rfunc(Y,Z),T),
	position_category(R,rfunc(lfunc(X,Y),Z),T),
	license_bwd_x_subst(Y).

%%%
%%% PUNCTUATION RULES
%%%

%
% left comma rule: , X => X
%
% action: rule_left_comma/3 (PositionLeft, PositionRight, Time)

precCat(rule_left_comma,L,T,Y):-actiontime(T),
						position_adjacent(L,R,T),
    position_category(L,",",T),
	position_category(R,Y,T),
	left_comma(Y).

%
% right comma rule:  X , => X
%
% action: rule_right_comma/3 (PositionLeft, PositionRight, Time)

precCat(rule_right_comma,L,T,Y):-actiontime(T),
						position_adjacent(L,R,T),
    position_category(L,Y,T),
	position_category(R,",",T),
	right_comma(Y).

%
% left colon rule: : X => X
%
% action: rule_left_colon/3 (PositionLeft, PositionRight, Time)

precCat(rule_left_colon,L,T,Y):-actiontime(T),
						position_adjacent(L,R,T),
    position_category(L,":",T),
	position_category(R,Y,T),
	left_colon(Y).

%
% right colon rule:  X : => X
%
% action: rule_right_colon/3 (PositionLeft, PositionRight, Time)

precCat(rule_right_colon,L,T,Y):-actiontime(T),
						position_adjacent(L,R,T),
    position_category(L,Y,T),
	position_category(R,":",T),
	right_colon(Y).

%
% left semicolon rule: ; X => X
%
% action: rule_left_semicolon/3 (PositionLeft, PositionRight, Time)

precCat(rule_left_semicolon,L,T,Y):-actiontime(T),
						position_adjacent(L,R,T),
    position_category(L,";",T),
	position_category(R,Y,T),
	left_colon(Y).

%
% right semicolon rule:  X , => X
%
% action: rule_right_semicolon/3 (PositionLeft, PositionRight, Time)

precCat(rule_right_semicolon,L,T,Y):-actiontime(T),
						position_adjacent(L,R,T),
    position_category(L,Y,T),
	position_category(R,";",T),
	right_colon(Y).



%%
%% Standard tenary combinator (Conj precCat predicate instantiaton 
%
% coordination rule: X cConj X => X
%
% action: rule_coordination/3 (PositionLeft, PositionRight, Time)
%
position_conj_adjacent(L,R,T):-	position_adjacent(L,M,T),
	position_category(M,CC,0),
    	conjunction_category(CC),
	position_adjacent(M,R,T).

precCat(rule_coordination,L,T,X):- 	actiontime(T),
	position_category(L,X,T),
	position_conj_adjacent(L,R,T),
	position_category(R,X,T).


% type raising limitations
% limitRaise/2 (FromCategory,ToCategory)
%   limits applicability of type raising:
%   to which category raising may apply,
%   and to which category may be raised.
%   if raising shall be possible to any category, use
%   "limitRaise(C1,C2) :- category(C1), category(C2)."
% TODO use better configuration from this, see emails and baldridge paper
% see www.ling.umd.edu/~alxndrw/CGReadings/steedman-dowty-1.pdf 3.3
limitRaise("NP","S"). % raise to S
limitRaise("NP",lfunc("S","NP")). % raise to VP = S\NP
limitRaise("NP",rfunc(lfunc("S","NP"),"NP")). % raise to TV = (S\NP)/NP
% type raising limitations
% limitRaiseFrom/1 (FromCategory)
limitRaiseFrom(From):-limitRaise(From,To).

%
% LICENSING/NORMALIZATION PRECONDITIONS
% 
 
% ban FA if left category was created by FC (we can use two FAs instead)
% example: X/Y Y/Z Z -> Z
:-occurs(rule_fwd_appl,L,T), actiontime(T), time(TLast),
 	position_last_affected(L,TLast,T), 
 	occurs(rule_fwd_comp,L,TLast1),t_tminus1(TLast,TLast1).

% ban FA if left category was created by FTR (we can use BA instead)
% example: A X\A
:-occurs(rule_fwd_appl,L,T), actiontime(T), time(TLast),
 	position_last_affected(L,TLast,T), 
 	occ(rule_fwd_raise,L,TLast1), t_tminus1(TLast,TLast1).

% ban BA if right category was created by BC (we can use two BAs instead)
% example: X Y\X Z\Y -> Z
:-occurs(rule_bwd_appl,L,T), actiontime(T), time(TLast),
  position_adjacent(L,R,T),
 	position_last_affected(R,TLast,T), 
 	occurs(rule_bwd_comp,R,TLast1), t_tminus1(TLast,TLast1).

% ban BA if right category was created by BTR (we can use FA instead)
% example: X/A A
:-occurs(rule_bwd_appl,L,T), actiontime(T), time(TLast),
  position_adjacent(L,R,T),
 	position_last_affected(R,TLast,T), 
 	occ(rule_bwd_raise,R,TLast1), t_tminus1(TLast,TLast1).

% ban BC if left category was created by BC (we can first do BC on the right and then on the left instead)
% example: Y\X Z\Y V\Z -> V\X
:-occurs(rule_bwd_comp,L,T), actiontime(T), time(TLast),
 	position_last_affected(L,TLast,T), 
 	occurs(rule_bwd_comp,L,TLast1), t_tminus1(TLast,TLast1).

% ban FC if left category was created by FC (we can first do FC on the right and then on the left instead)
% example: X/Y Y/Z Z/V -> X/V
:-occurs(rule_fwd_comp,L,T), actiontime(T), time(TLast),
 	position_last_affected(L,TLast,T), 
 	occurs(rule_fwd_comp,L,TLast1), t_tminus1(TLast,TLast1).

% ban coordination if left and right category was raised to the same category (we can raise afterwards)
% example: Y conj Y -> X/(X\Y)
% example: Y conj Y -> X\(X/Y)
raisingCombinator(rule_fwd_raise;rule_bwd_raise).
:-occurs(rule_coordination,L,T), actiontime(T), time(TLastL), time(TLastR),
	position_conj_adjacent(L,R,T),
 	position_affected(L,T), 
 	position_last_affected(R,TLastR,T), 
 	occurs(Raise,L,X,T1), t_tminus1(T,T1),
 	occurs(Raise,R,X,TLastR1), t_tminus1(TLastR,TLastR1),
  raisingCombinator(Raise).

:-occurs(rule_coordination,L,T), actiontime(T), time(TLastL), time(TLastR),
	position_conj_adjacent(L,R,T),
 	position_last_affected(L,TLastL,T), 
 	position_affected(R,T), 
 	occurs(Raise,L,X,TLastL1), t_tminus1(TLastL,TLastL1),
 	occurs(Raise,R,X,T1), t_tminus1(T,T1),
  raisingCombinator(Raise).

unary_type_change("N","NP").
unary_type_change("NP",rfunc("S",rfunc("S","NP"))).
unary_type_change("NP",rfunc("NP",lfunc("NP","NP"))).


left_comma("N").
left_comma("NP").
left_comma("S").
left_comma(rfunc("N","N")).
left_comma(lfunc("NP","NP")).
left_comma(lfunc("PP","PP")).
left_comma(rfunc("S","S")).
left_comma(lfunc("S","S")).
left_comma(lfunc("S","NP")).
left_comma(lfunc(lfunc("S","NP"),lfunc("S","NP"))).
left_comma(rfunc(lfunc("S","NP"),lfunc("S","NP"))).
left_comma(lfunc(lfunc(lfunc("S","NP"),lfunc("S","NP")),lfunc(lfunc("S","NP"),lfunc("S","NP")))).

right_comma("N").
right_comma("NP").
right_comma("PP").
right_comma("S").
right_comma(rfunc("N","N")).
right_comma(lfunc("NP","NP")).
right_comma(rfunc("S","S")).
right_comma(lfunc("S","S")).
right_comma(lfunc("S","NP")).
right_comma(rfunc(lfunc("S","NP"),"S")).
right_comma(lfunc(lfunc("S","S"),"NP")).
right_comma(rfunc(lfunc("S","NP"),"NP")).
right_comma(rfunc(lfunc("S","NP"),"PP")).
right_comma(rfunc(lfunc("NP","NP"),lfunc("S","NP"))).
right_comma(lfunc(lfunc("S","NP"),lfunc("S","NP"))).
right_comma(rfunc(lfunc("S","NP"),lfunc("S","NP"))).

left_colon("N").
left_colon("NP").
left_colon("S").
left_colon(rfunc("N","N")).
left_colon(lfunc("NP","NP")).
left_colon(lfunc("S","NP")).
left_colon(lfunc(lfunc("S","NP"),lfunc("S","NP"))).

right_colon("N").
right_colon("NP").
right_colon("PP").
right_colon("S").
right_colon(lfunc("NP","NP")).
right_colon(rfunc("S","S")).
right_colon(lfunc("S","NP")).
right_colon(rfunc(lfunc("S","NP"),"S")).
right_colon(lfunc(lfunc("S","NP"),lfunc("S","NP"))).
right_colon(rfunc(lfunc("S","NP"),lfunc("S","NP"))).


left_bracket("N").
left_bracket("NP").
left_bracket("S").
left_bracket(lfunc("NP","NP")).
left_bracket(lfunc(lfunc("S","NP"),lfunc("S","NP"))).

right_bracket("N").
right_bracket("NP").
right_bracket("S").
right_bracket(lfunc("N","N")).
right_bracket(rfunc("N","N")).
right_bracket(lfunc("NP","NP")).
right_bracket(lfunc("S","NP")).
right_bracket(rfunc("S","S")).
right_bracket(lfunc("S","S")).
right_bracket(lfunc(rfunc("N","N"),rfunc("N","N"))).
right_bracket(lfunc(lfunc("S","NP"),lfunc("S","NP"))).
right_bracket(rfunc(lfunc("S","NP"),lfunc("S","NP"))).
